"""Test."""
import unittest

from utils_base import Parallel


class TestMR(unittest.TestCase):
    """Test."""

    def test_map_parallel(self):
        """Test."""

        def func_map(x):
            return x**3

        input_list = list(range(0, 10))
        expected_output = list(
            map(
                func_map,
                input_list,
            )
        )
        actual_output = Parallel.map(
            func_map,
            input_list,
            max_threads=4,
        )
        self.assertEqual(
            expected_output,
            actual_output,
        )



import os
from functools import cached_property

import camelot
import PyPDF2

from utils_base.file.Directory import Directory
from utils_base.file.File import File


class PDFFile(File):
    @cached_property
    def n_pages(self):
        file = open(self.path, 'rb')
        reader = PyPDF2.PdfReader(file)
        return len(reader.pages)

    @property
    def dir_tables(self):
        return Directory(self.path + '.tables')

    def load_table_files(self):
        return self.dir_tables.children

    def store_table_files(self):
        os.makedirs(self.dir_tables.path)
        tables = camelot.read_pdf(self.path, pages='all')
        table_files = []
        for i, table in enumerate(tables):
            table_path = os.path.join(self.dir_tables.path, f'{i}.tsv')
            table.to_csv(table_path, sep='\t')
            table_files.append(File(table_path))
        return table_files

    @property
    def table_files(self):
        if self.dir_tables.exists:
            return self.load_table_files()
        return self.store_table_files()



"""Hashx."""
import hashlib


class Hash:
    @staticmethod
    def md5(s):
        """Compute MD5 Hash."""
        _md5 = hashlib.md5()
        _md5.update(s.encode())
        return _md5.hexdigest()



import math

import geopy.distance

from utils_base.ds import Parse


class LatLng:
    def __init__(self, lat: float, lng: float):
        self.lat = lat
        self.lng = lng

    def __eq__(self, other):
        return self.lat == other.lat and self.lng == other.lng

    def __hash__(self):
        return hash((self.lat, self.lng))

    def __str__(self):
        lat_str = f'{abs(self.lat):.6f}°' + ('N' if self.lat > 0 else 'S')
        lng_str = f'{abs(self.lng):.6f}°' + ('E' if self.lng > 0 else 'W')
        return f'{lat_str}, {lng_str}'

    def __repr__(self):
        return str(self)

    @staticmethod
    def parse(latlng_str_original: str) -> str:
        latlng_str = latlng_str_original.replace('°', ' ').strip()
        for x in ['N', 'S', 'E', 'W']:
            latlng_str = latlng_str.replace(x, '')
        lat_str, lng_str = latlng_str.split(',')
        lat = Parse.float(lat_str) * (-1 if 'S' in latlng_str_original else 1)
        lng = Parse.float(lng_str) * (-1 if 'W' in latlng_str_original else 1)
        return LatLng(lat, lng)

    def distance(self, other: 'LatLng') -> float:
        return geopy.distance.geodesic(
            (self.lat, self.lng), (other.lat, other.lng)
        ).km

    def angle(self, other: 'LatLng') -> float:
        return (
            math.atan2(other.lat - self.lat, other.lng - self.lng)
            * 180
            / math.pi
        )

    @staticmethod
    def bbox(latlng_list: list['LatLng']) -> tuple['LatLng']:
        lat_list = [latlng.lat for latlng in latlng_list]
        lng_list = [latlng.lng for latlng in latlng_list]
        return (
            LatLng(min(lat_list), min(lng_list)),
            LatLng(max(lat_list), max(lng_list)),
        )

    @staticmethod
    def get_func_t(
        latlng_list: list['LatLng'],
        width: float,
        height: float,
        padding: float,
    ) -> callable:
        width_actual = width - padding * 2
        height_actual = height - padding * 2

        (min_latlng, max_latlng) = LatLng.bbox(latlng_list)

        def t(latlng):
            x = (latlng.lng - min_latlng.lng) / (
                max_latlng.lng - min_latlng.lng
            )
            y = (latlng.lat - min_latlng.lat) / (
                max_latlng.lat - min_latlng.lat
            )
            return (
                x * width_actual + padding,
                (1 - y) * height_actual + padding,
            )

        return t



# utils_base.ds (auto generate by build_inits.py)

from utils_base.ds.Dict import Dict
from utils_base.ds.Iter import Iter
from utils_base.ds.List import List
from utils_base.ds.Parse import Parse
from utils_base.ds.String import String



# utils_base.file (auto generate by build_inits.py)

from utils_base.file.CSVFile import CSVFile
from utils_base.file.Directory import Directory
from utils_base.file.File import File
from utils_base.file.FiledVariable import FiledVariable
from utils_base.file.FileOrDirectory import FileOrDirectory
from utils_base.file.JSONFile import JSONFile
from utils_base.file.PDFFile import PDFFile
from utils_base.file.TSVFile import TSVFile
from utils_base.file.XSVFile import XSVFile
from utils_base.file.Zip import Zip



import os
import zipfile


class Zip:  # noqa
    def __init__(self, path):
        self.path = path

    @property
    def zip_path(self):
        return self.path + '.zip'

    @property
    def arc_name(self):
        return os.path.basename(os.path.normpath(self.path))

    @property
    def dir_zip(self):
        return os.path.dirname(os.path.normpath(self.path))

    def zip(self, skip_delete=False):
        assert os.path.exists(self.path)
        with zipfile.ZipFile(
            self.zip_path,
            mode='w',
            compression=zipfile.ZIP_DEFLATED,
        ) as zip_file:
            zip_file.write(self.path, arcname=self.arc_name)
            assert os.path.exists(self.zip_path)

        if not skip_delete:
            os.remove(self.path)
            assert not os.path.exists(self.path)

    def unzip(self, skip_delete=False):  # noqa
        assert os.path.exists(self.zip_path)
        with zipfile.ZipFile(self.zip_path) as zip_file:
            zip_file.extractall(self.dir_zip)
            assert os.path.exists(self.path)

        if not skip_delete:
            os.remove(self.zip_path)
            assert not os.path.exists(self.zip_path)



class TimeDelta:
    def __init__(self, dut: int = 0):
        self.dut = dut

    def __eq__(self, other) -> bool:
        return self.dut == other.dut

    def humanize(self):
        dut = self.dut

        for dut_limit, label in [
            (60, 'second'),
            (60, 'minute'),
            (24, 'hour'),
            (7, 'day'),
            (None, 'week'),
        ]:
            if not dut_limit or dut < dut_limit:
                suffix = ''
                if dut != 1:
                    suffix = 's'
                return f'{dut:.0f} {label}{suffix}'
            else:
                dut /= dut_limit



import xml.dom.minidom as minidom
import xml.etree.ElementTree as ElementTree

from utils_base.file.File import File

FONT_FAMILY = 'sans-serif'
DEFAULT_ATTRIB_MAP = {
    'html': {
        'style': 'font-family: %s;' % FONT_FAMILY,
    },
    'svg': {
        'xmlns': 'http://www.w3.org/2000/svg',
    },
}


class XMLElement:
    def __init__(
        self,
        tag: str,
        child_list_or_str_or_other=None,
        attrib_custom={},
    ):
        self.tag = tag
        self.child_list_or_str_or_other = child_list_or_str_or_other
        self.attrib_custom = attrib_custom

    @property
    def expanded_attrib(self):
        expanded_attrib = DEFAULT_ATTRIB_MAP.get(self.tag, {})
        expanded_attrib.update(self.attrib_custom)
        expanded_attrib = dict(
            zip(
                list(
                    map(
                        lambda k: k.replace('_', '-'),
                        expanded_attrib.keys(),
                    )
                ),
                list(map(str, expanded_attrib.values())),
            ),
        )
        return expanded_attrib

    @property
    def tag_real(self):
        return self.tag.split('-')[0]

    @property
    def child_element_list(self):
        if not isinstance(self.child_list_or_str_or_other, list):
            return []
        nonnull_child_list = list(
            filter(
                lambda child_or_none: child_or_none is not None,
                self.child_list_or_str_or_other,
            )
        )
        return list(
            map(
                lambda child: child.element,
                nonnull_child_list,
            )
        )

    @property
    def text(self):
        if isinstance(self.child_list_or_str_or_other, str):
            return self.child_list_or_str_or_other
        return ''

    @property
    def element(self):
        element = ElementTree.Element(self.tag_real)
        element.attrib = self.expanded_attrib  # noqa
        for child_element in self.child_element_list:
            element.append(child_element)
        element.text = self.text
        return element

    def __str__(self):
        s = ElementTree.tostring(self.element, encoding='utf-8').decode()
        parsed_s = minidom.parseString(s)
        return parsed_s.toprettyxml(indent='  ')

    def __repr__(self):
        return self.__str__()

    def store(self, xml_file):
        File(xml_file).write(str(self))



class Parse:
    @staticmethod
    def _clean_(x: str) -> str:
        return x.strip().lower().replace(',', '')

    @staticmethod
    def int(x) -> int:
        try:
            return int(Parse._clean_(x))
        except ValueError:
            return None

    @staticmethod
    def float(x) -> float:
        try:
            return float(Parse._clean_(x))
        except ValueError:
            return None



import time

from utils_base.time.Time import Time
from utils_base.time.TimeZoneOffset import TimeZoneOffset


class TimeFormat:
    def __init__(self, format_str: str, timezone_offset=TimeZoneOffset.LK):
        self.format_str = format_str
        self.timezone_offset = timezone_offset

    @property
    def dut_timezone(self):
        return time.timezone - self.timezone_offset

    def parse(self, time_str: str) -> Time:
        ut_base = time.mktime(time.strptime(time_str, self.format_str))
        ut = ut_base - self.dut_timezone
        return Time(ut)

    def stringify(self, t: Time) -> str:
        return time.strftime(
            self.format_str, time.localtime(t.ut + self.dut_timezone)
        )

    def format(self, t: Time) -> str:
        return self.stringify(t)

    @property
    def formatNow(self) -> str:
        return self.format(Time.now())


TimeFormat.DATE = TimeFormat('%Y-%m-%d')  # noqa
TimeFormat.TIME = TimeFormat('%Y-%m-%d %H:%M:%S')  # noqa

TimeFormat.DATE_ID = TimeFormat('%Y%m%d')  # noqa
TimeFormat.TIME_ID = TimeFormat('%Y%m%d.%H%M%S')  # noqa



from utils_base.geo.LatLng import LatLng


class LatLngLK(LatLng):
    # Cities
    COLOMBO = LatLng(6.917273734591578, 79.86479237575594)
    KANDY = LatLng(7.293495849930314, 80.64101416983969)
    GALLE = LatLng(6.029591949643258, 80.2161060278968)
    JAFFNA = LatLng(9.665142806084814, 80.00933289927613)

    # Extreme Points
    NORTH = LatLng(9.835556, 80.212222)
    SOUTH = LatLng(5.923389, 80.589694)
    EAST = LatLng(7.022222, 81.879167)
    WEST = LatLng(9.383333, 79.516667)

    BBOX = LatLng.bbox([NORTH, SOUTH, EAST, WEST])

    @staticmethod
    def get_func_t_lk(
        width: float,
        height: float,
        padding: float,
    ):
        return LatLng.get_func_t(
            [
                LatLngLK.NORTH,
                LatLngLK.SOUTH,
                LatLngLK.EAST,
                LatLngLK.WEST,
            ],
            width,
            height,
            padding,
        )



# utils_base.time (auto generate by build_inits.py)

from utils_base.time.Time import Time
from utils_base.time.TimeDelta import TimeDelta
from utils_base.time.TimeFormat import TimeFormat
from utils_base.time.TimeUnit import TimeUnit
from utils_base.time.TimeZoneOffset import TimeZoneOffset



"""Utils."""

import logging

from utils_base.console.Console import Console
from utils_base.console.constants import LEVEL_TO_STYLE


class CustomLoggingFormatter(logging.Formatter):
    def format(self, record):
        style = LEVEL_TO_STYLE[record.levelno]
        return Console.format(
            (f'[{record.name}]' if record.name else ''),
            str(record.msg),
            **style,
        )


class Log(logging.Logger):
    def __init__(self, name: str, level: int = logging.DEBUG):
        super(Log, self).__init__(name, level)
        self.propagate = False

        formatter = CustomLoggingFormatter()
        sh = logging.StreamHandler()
        sh.setLevel(logging.DEBUG)
        sh.setFormatter(formatter)
        self.handlers = [sh]  # noqa


Log.default = Log('default')
Log.main = Log('main')
Log.pipeline = Log('pipeline')



import logging

from utils_base.console.constants.COLOR_BACKGROUND import COLOR_BACKGROUND
from utils_base.console.constants.COLOR_FOREGROUND import COLOR_FOREGROUND
from utils_base.console.constants.COLOR_FORMAT import COLOR_FORMAT

LEVEL_TO_STYLE = {
    logging.CRITICAL: dict(
        foreground=COLOR_FOREGROUND.WHITE,
        background=COLOR_BACKGROUND.RED,
        format=COLOR_FORMAT.BOLD,
    ),
    logging.ERROR: dict(
        foreground=COLOR_FOREGROUND.RED,
        background=COLOR_BACKGROUND.BLACK,
    ),
    logging.WARNING: dict(
        foreground=COLOR_FOREGROUND.YELLOW,
        background=COLOR_BACKGROUND.BLACK,
    ),
    logging.INFO: dict(
        foreground=COLOR_FOREGROUND.GREEN,
        background=COLOR_BACKGROUND.BLACK,
    ),
    logging.DEBUG: dict(
        foreground=COLOR_FOREGROUND.WHITE,
        background=COLOR_BACKGROUND.BLACK,
        format=COLOR_FORMAT.FAINT,
    ),
    logging.NOTSET: dict(
        foreground=COLOR_FOREGROUND.WHITE,
        background=COLOR_BACKGROUND.BLACK,
    ),
}



import os
import tempfile

from PIL import Image as PImage
from PIL import ImageDraw, ImageEnhance, ImageFont

from utils_base.console.Log import Log

log = Log(__name__)


class DEFAULT:
    FONT_FAMILY = os.path.join('src', 'utils_base', 'image', 'trebuc.ttf')
    FONT_SIZE = 40
    FILL = (0, 0, 0)


class Image:
    def __init__(self, im: PImage):
        self.im = im

    @staticmethod
    def load(image_path: str):  # -> Image
        return Image(PImage.open(image_path))

    def write(self, image_path: str) -> str:
        self.im.save(image_path)
        log.debug(f'Saved {image_path}')
        return image_path

    def write_temp(self) -> str:
        return self.write(tempfile.NamedTemporaryFile(suffix='.png').name)

    @property
    def size(self):
        return self.im.size

    def crop(
        self, left_top: tuple[float], width_height: tuple[float]
    ):  # -> Image
        width, height = width_height
        im_width, im_height = self.im.size

        if height > im_height or width > im_width:
            return self

        left, top = left_top
        bbox = (left, top, left + width, top + height)
        im = self.im.crop(bbox)
        log.debug(f'crop: {left},{top} & {width}x{height}')
        return Image(im)

    def resize(self, ratio: float):  # -> Image
        im = self.im
        width, height = im.size
        new_width, new_height = (int(width * ratio), int(height * ratio))
        im = im.resize((new_width, new_height))
        log.debug(f'resize({ratio}) -> {new_width}x{new_height}')
        return Image(im)

    def draw_text(
        self,
        left_top: tuple[float],
        text: str,
        fill: tuple = None,
        font_family: str = None,
        font_size: int = None,
    ):  # -> Image
        draw = ImageDraw.Draw(self.im)
        font = ImageFont.truetype(
            font_family or DEFAULT.FONT_FAMILY, font_size or DEFAULT.FONT_SIZE
        )
        draw.text(left_top, text, fill=fill or DEFAULT.FILL, font=font)
        left, top = left_top
        log.debug(f'draw_text(({left},{top}), "{text}")')
        return Image(self.im)

    @staticmethod
    def equalize_map(x_list: list[float]) -> list[float]:
        x_list_sorted = sorted(x_list)
        n = len(x_list_sorted)
        idx = {}
        for i, x in enumerate(x_list_sorted):
            p = i / n
            if x not in idx:
                idx[x] = p
        return idx

    @staticmethod
    def equalize(c, min_v: float, max_v: float):
        assert min_v < max_v
        idx = Image.equalize_map(list(c.getdata()))
        span = max_v - min_v
        return c.point(lambda x: int(min_v + idx.get(x, 0) * span))

    def equalize_hue(self, min_v: float = 0, max_v: float = 360) -> 'Image':
        im = self.im.convert('HSV')
        h, s, v = im.split()
        h = Image.equalize(h, min_v, max_v)

        im = PImage.merge('HSV', (h, s, v))
        im = im.convert('RGB')
        return Image(im)

    def enhance(self, factor: float) -> 'Image':
        enhancer = ImageEnhance.Contrast(self.im)
        im2 = enhancer.enhance(factor)
        return Image(im2)



class Iter:
    def count(self, func_key=None):
        key_to_n = {}
        for item in self:
            key = func_key(item) if func_key else item
            if key not in key_to_n:
                key_to_n[key] = 0
            key_to_n[key] += 1
        return key_to_n



from utils_base.console.constants import (COLOR_BACKGROUND, COLOR_FOREGROUND,
                                          COLOR_FORMAT)


class Console:
    @staticmethod
    def format(*args, **kwargs) -> str:
        foreground = kwargs.get('foreground', '')
        background = kwargs.get('background', '')
        format = kwargs.get('format', '')
        text = ' '.join(args)
        return ''.join(
            [foreground, background, format, text, COLOR_FORMAT.RESET]
        )

    @staticmethod
    def print(*args, **kwargs):
        print(Console.format(*args, **kwargs))

    @staticmethod
    def h1(*args) -> str:
        return Console.format(
            *args,
            foreground=COLOR_FOREGROUND.WHITE,
            background=COLOR_BACKGROUND.GREEN,
            format=COLOR_FORMAT.UNDERLINE,
        )

    @staticmethod
    def h2(*args) -> str:
        return Console.format(
            *args,
            foreground=COLOR_FOREGROUND.GREEN,
        )

    @staticmethod
    def normal(*args) -> str:
        return Console.format(*args, foreground=COLOR_FOREGROUND.WHITE)

    @staticmethod
    def note(*args) -> str:
        return Console.format(
            *args,
            foreground=COLOR_FOREGROUND.YELLOW,
            format=COLOR_FORMAT.ITALIC,
        )

    @staticmethod
    def md5_line(line) -> str:
        if not line:
            return ''

        if line.startswith('# '):
            return Console.h1(line[2:])
        if line.startswith('## '):
            return Console.h2(line[3:])
        if line.startswith('*') and line.endswith('*'):
            return Console.note(line[1:-1])
        return Console.normal(line)

    @staticmethod
    def md5(*args) -> str:
        return '\n'.join(Console.md5_line(arg) for arg in args)

    @staticmethod
    def print_lines(*args):
        for arg in args:
            print(arg)



# utils_base (auto generate by build_inits.py)

from utils_base.console import (COLOR_BACKGROUND, COLOR_FOREGROUND,
                                COLOR_FORMAT, LEVEL_TO_STYLE, Console, Log)
from utils_base.ds import Dict, Iter, List, Parse, String
from utils_base.file import (CSVFile, Directory, File, FiledVariable,
                             FileOrDirectory, JSONFile, PDFFile, TSVFile,
                             XSVFile, Zip)
from utils_base.geo import LatLng, LatLngLK
from utils_base.Hash import Hash
from utils_base.image import Image
from utils_base.Parallel import Parallel
from utils_base.time import (Time, TimeDelta, TimeFormat, TimeUnit,
                             TimeZoneOffset)
from utils_base.xml import XMLElement, XMLUtils, _



# utils_base.console.constants (auto generate by build_inits.py)

from utils_base.console.constants.COLOR_BACKGROUND import COLOR_BACKGROUND
from utils_base.console.constants.COLOR_FOREGROUND import COLOR_FOREGROUND
from utils_base.console.constants.COLOR_FORMAT import COLOR_FORMAT
from utils_base.console.constants.LEVEL_TO_STYLE import LEVEL_TO_STYLE



class TimeUnit:
    def __init__(self, seconds: int):
        self.seconds = seconds

    def __truediv__(self, other):
        if isinstance(other, TimeUnit):
            return self.seconds / other.seconds
        if isinstance(other, float) or isinstance(other, int):
            return TimeUnit(self.seconds / other)
        raise TypeError(
            f'unsupported operand type(s) for /: {type(self)} and {type(other)}'
        )

    def __mul__(self, other):
        return TimeUnit(self.seconds * other)

    def __eq__(self, other):
        if isinstance(other, TimeUnit):
            return self.seconds == other.seconds
        return False


TimeUnit.SECOND = TimeUnit(1)
TimeUnit.MINUTE = TimeUnit.SECOND * 60
TimeUnit.HOUR = TimeUnit.MINUTE * 60
TimeUnit.DAY = TimeUnit.HOUR * 24
TimeUnit.WEEK = TimeUnit.DAY * 7
TimeUnit.FORTNIGHT = TimeUnit.WEEK * 7

TimeUnit.AVG_YEAR = TimeUnit.DAY * 365.25
TimeUnit.AVG_QTR = TimeUnit.AVG_YEAR / 4
TimeUnit.AVG_MONTH = TimeUnit.AVG_YEAR / 12


class SECONDS_IN:  # noqa
    MINUTE = TimeUnit.MINUTE / TimeUnit.SECOND
    HOUR = TimeUnit.HOUR / TimeUnit.SECOND
    DAY = TimeUnit.DAY / TimeUnit.SECOND
    WEEK = TimeUnit.WEEK / TimeUnit.SECOND
    FORTNIGHT = TimeUnit.FORTNIGHT / TimeUnit.SECOND

    AVG_MONTH = TimeUnit.AVG_MONTH / TimeUnit.SECOND
    AVG_QTR = TimeUnit.AVG_QTR / TimeUnit.SECOND
    AVG_YEAR = TimeUnit.AVG_YEAR / TimeUnit.SECOND


TimeUnit.SECOND_IN = SECONDS_IN


class DAYS_IN:  # noqa
    AVG_MONTH = TimeUnit.AVG_MONTH / TimeUnit.DAY
    AVG_QTR = TimeUnit.AVG_QTR / TimeUnit.DAY
    AVG_YEAR = TimeUnit.AVG_YEAR / TimeUnit.DAY


TimeUnit.DAYS_IN = DAYS_IN



from utils_base.ds.Iter import Iter


class Dict(Iter):
    def __init__(self, x={}):
        assert isinstance(x, dict)
        self.x = x

    def todict(self):
        return self.x

    def keys(self):
        return self.x.keys()

    def values(self):
        return self.x.values()

    def items(self):
        return self.x.items()

    def items_sorted_by_key(self):
        return sorted(
            self.x.items(),
            key=lambda item: item[0],
        )

    def items_sorted_by_value(self):
        return sorted(
            self.x.items(),
            key=lambda item: item[1],
        )

    def len(self):
        return len(self.x)

    def __eq__(self, other):
        if isinstance(other, Dict):
            return self.x == other.x
        if isinstance(other, dict):
            return self.x == other
        return False

    def __getitem__(self, key):
        return self.x[key]

    def __setitem__(self, key, value):
        self.x[key] = value

    def __delitem__(self, key):
        del self.x[key]

    def __iter__(self):
        return self.x.__iter__()

    def extract_keys(self, keys):
        return Dict(
            dict(
                list(
                    filter(
                        lambda item: item[0] in keys,
                        self.x.items(),
                    )
                )
            )
        )

    def __str__(self):
        return self.x.__str__()

    def __repr__(self):
        return self.x.__repr__()



class COLOR_FOREGROUND:
    BLACK = '\033[30m'  # noqa
    RED = '\033[31m'  # noqa
    GREEN = '\033[32m'  # noqa
    YELLOW = '\033[33m'  # noqa
    BLUE = '\033[34m'  # noqa
    MAGENTA = '\033[35m'  # noqa
    CYAN = '\033[36m'  # noqa
    LIGHT_GRAY = '\033[37m'  # noqa
    GRAY = '\033[90m'  # noqa
    LIGHT_RED = '\033[91m'  # noqa
    LIGHT_GREEN = '\033[92m'  # noqa
    LIGHT_YELLOW = '\033[93m'  # noqa
    LIGHT_BLUE = '\033[94m'  # noqa
    LIGHT_MAGENTA = '\033[95m'  # noqa
    LIGHT_CYAN = '\033[96m'  # noqa
    WHITE = '\033[97m'  # noqa



from utils_base.ds.String import String
from utils_base.xml._ import _

FONT_FAMILY = 'sans-serif'
DEFAULT_ATTRIB_MAP = {
    'html': {
        'style': 'font-family: %s;' % FONT_FAMILY,
    },
    'svg': {
        'xmlns': 'http://www.w3.org/2000/svg',
    },
}


class XMLUtils:
    @staticmethod
    def render_link_styles(css_file='styles.css'):
        return _('link', None, {'rel': 'stylesheet', 'href': css_file})

    @staticmethod
    def style(**kwargs):
        style_content = ''.join(
            list(
                map(
                    lambda item: '%s:%s;'
                    % (String(str(item[0])).kebab, str(item[1])),
                    kwargs.items(),
                )
            )
        )
        return dict(style=style_content)



class COLOR_BACKGROUND:  # noqa
    BLACK = '\033[40m'  # noqa
    RED = '\033[41m'  # noqa
    GREEN = '\033[42m'  # noqa
    YELLOW = '\033[43m'  # noqa
    BLUE = '\033[44m'  # noqa
    MAGENTA = '\033[45m'  # noqa
    CYAN = '\033[46m'  # noqa
    LIGHT_GRAY = '\033[47m'  # noqa
    GRAY = '\033[100m'  # noqa
    LIGHT_RED = '\033[101m'  # noqa
    LIGHT_GREEN = '\033[102m'  # noqa
    LIGHT_YELLOW = '\033[103m'  # noqa
    LIGHT_BLUE = '\033[104m'  # noqa
    LIGHT_MAGENTA = '\033[105m'  # noqa
    LIGHT_CYAN = '\033[106m'  # noqa
    WHITE = '\033[107m'  # noqa



"""Map-Reduce utils_base."""

import concurrent.futures

DEFAULT_MAX_THREADS = 4


class Parallel:
    @staticmethod
    def run(workers, max_threads=DEFAULT_MAX_THREADS):
        """Run workers in parallel."""
        with concurrent.futures.ThreadPoolExecutor(
            max_threads
        ) as thread_pool:
            future_list = []
            for worker in workers:
                future = thread_pool.submit(worker)
                future_list.append(future)

            for future in future_list:
                future.done()

            output_list = []
            for future in future_list:
                output_list.append(future.result())
            return output_list

    @staticmethod
    def map(func_map, params_list, max_threads=DEFAULT_MAX_THREADS):
        """Run list(map(...)) in parallel.

        Args:
            func_map (function): Mapper function
            params_list (list): Params to be mapped
            max_threads (int, optional): Maximum parallel threads.
                DEFAULT_MAX_THREADS = 4

        .. code-block:: python

            >>> from utils_base import mr
            >>> print(mr.map_parallel(lambda x: x ** 2, [1, 2, 3, 4]))
            [1, 4, 9, 16]

        """

        def get_worker(params):
            def worker():
                return func_map(params)

            return worker

        workers = list(
            map(
                lambda params: get_worker(params=params),
                params_list,
            )
        )
        return Parallel.run(workers, max_threads)



"""Setup."""

import setuptools

DIST_NAME = 'utils_base'
VERSION = "1.0.9"
DESCRIPTION = "Utilities that extend Standard Python."
INSTALL_REQUIRES = [
    'camelot-py',
    'opencv-python',
    'PyPDF2==2.0.0',
    'Pillow',
    'geopy',
]

setuptools.setup(
    name="%s-nuuuwan" % DIST_NAME,
    version=VERSION,
    author="Nuwan I. Senaratna",
    author_email="nuuuwan@gmail.com",
    description=DESCRIPTION,
    long_description="",
    long_description_content_type="text/markdown",
    url="https://github.com/nuuuwan/%s" % DIST_NAME,
    project_urls={
        "Bug Tracker": "https://github.com/nuuuwan/%s/issues" % DIST_NAME,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.10",
    install_requires=INSTALL_REQUIRES,
    test_suite='nose.collector',
    tests_require=['nose'],
)



class COLOR_FORMAT:
    RESET = '\033[0m'  # noqa
    NORMAL = ''  # noqa
    BOLD = '\033[01m'  # noqa
    FAINT = '\033[02m'  # noqa
    ITALIC = '\033[03m'  # noqa
    UNDERLINE = '\033[04m'  # noqa

    REVERSE = '\033[07m'  # noqa
    INVISIBLE = '\033[08m'  # noqa
    STRIKETHROUGH = '\033[09m'  # noqa



import tempfile

from utils_base.file.JSONFile import JSONFile
from utils_base.Hash import Hash

FILE_VARIABLE_CACHE = {}


class FiledVariable:
    def __init__(self, key: str, func_get):
        self.key = key
        self.func_get = func_get

    @property
    def cache_key(self):
        return Hash.md5(self.key)

    @property
    def file_path(self):
        return tempfile.NamedTemporaryFile(
            prefix="cache.", suffix=".json"
        ).name

    @property
    def file(self):
        return JSONFile(self.file_path)

    def clear_file(self):
        self.file.delete()

    def clear_cache(self):
        global FILE_VARIABLE_CACHE
        FILE_VARIABLE_CACHE = {}

    @property
    def value(self):
        global FILE_VARIABLE_CACHE
        if self.cache_key in FILE_VARIABLE_CACHE:
            return FILE_VARIABLE_CACHE[self.cache_key]

        if self.file.exists:
            return self.file.read()

        value = self.func_get()
        self.file.write(value)
        FILE_VARIABLE_CACHE[self.cache_key] = value
        return value



# utils_base.console (auto generate by build_inits.py)

from utils_base.console.Console import Console
from utils_base.console.constants import (COLOR_BACKGROUND, COLOR_FOREGROUND,
                                          COLOR_FORMAT, LEVEL_TO_STYLE)
from utils_base.console.Log import Log



import time

from utils_base.time.TimeDelta import TimeDelta


class Time:
    def __init__(self, ut: int = None):
        if ut is None:
            ut = time.time()
        self.ut = ut

    def __eq__(self, other) -> bool:
        return self.ut == other.ut

    def __sub__(self, other) -> TimeDelta:
        if isinstance(other, TimeDelta):
            return Time(self.ut - other.dut)
        if isinstance(other, Time):
            return TimeDelta(self.ut - other.ut)
        raise TypeError(
            "unsupported operand type(s) for -: 'Time' and '%s'" % type(other)
        )

    def __add__(self, other):
        if isinstance(other, TimeDelta):
            return Time(self.ut + other.dut)
        raise TypeError(
            "unsupported operand type(s) for -: 'Time' and '%s'" % type(other)
        )

    @staticmethod
    def now():
        return Time()
